@@ -1,6 +1,7 @@
 {-# LANGUAGE CPP       #-}
 {-# LANGUAGE MagicHash #-}
 {-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE BangPatterns #-}
 
 -- |
 -- Module      : Codec.CBOR.FlatTerm
@@ -11,16 +12,16 @@
 -- Stability   : experimental
 -- Portability : non-portable (GHC extensions)
 --
--- A simpler form than CBOR for writing out @'Enc.Encoding'@ values that allows
+-- A simpler form than CBOR for writing out 'Enc.Encoding' values that allows
 -- easier verification and testing. While this library primarily focuses
--- on taking @'Enc.Encoding'@ values (independent of any underlying format)
+-- on taking 'Enc.Encoding' values (independent of any underlying format)
 -- and serializing them into CBOR format, this module offers an alternative
--- format called @'FlatTerm'@ for serializing @'Enc.Encoding'@ values.
+-- format called 'FlatTerm' for serializing 'Enc.Encoding' values.
 --
--- The @'FlatTerm'@ form is very simple and internally mirrors the original
--- @'Encoding'@ type very carefully. The intention here is that once you
--- have @'Enc.Encoding'@ and @'Dec.Decoding'@ values for your types, you can
--- round-trip values through @'FlatTerm'@ to catch bugs more easily and with
+-- The 'FlatTerm' form is very simple and internally mirrors the original
+-- 'Encoding' type very carefully. The intention here is that once you
+-- have 'Enc.Encoding' and 'Dec.Decoding' values for your types, you can
+-- round-trip values through 'FlatTerm' to catch bugs more easily and with
 -- a smaller amount of code to look through.
 --
 -- For that reason, this module is primarily useful for client libraries,
@@ -36,6 +37,7 @@ module Codec.CBOR.FlatTerm
   , toFlatTerm    -- :: Encoding -> FlatTerm
   , fromFlatTerm  -- :: Decoder s a -> FlatTerm -> Either String a
   , validFlatTerm -- :: FlatTerm -> Bool
+  , decodeTermToken -- Decoder s TermToken
   ) where
 
 #include "cbor.h"
@@ -43,6 +45,7 @@ module Codec.CBOR.FlatTerm
 import           Codec.CBOR.Encoding (Encoding(..))
 import qualified Codec.CBOR.Encoding as Enc
 import           Codec.CBOR.Decoding as Dec
+import qualified Codec.CBOR.Read     as Read
 import qualified Codec.CBOR.ByteArray        as BA
 import qualified Codec.CBOR.ByteArray.Sliced as BAS
 
@@ -60,19 +63,23 @@ import           Data.Word
 import           Data.Text (Text)
 import qualified Data.Text.Encoding as TE
 import           Data.ByteString (ByteString)
+import qualified Data.ByteString as BS
 import           Control.Monad.ST
+import qualified Control.Monad.ST.Lazy as ST.Lazy
+
+import Prelude hiding (encodeFloat, decodeFloat)
 
 
 --------------------------------------------------------------------------------
 
--- | A "flat" representation of an @'Enc.Encoding'@ value,
+-- | A \"flat\" representation of an 'Enc.Encoding' value,
 -- useful for round-tripping and writing tests.
 --
 -- @since 0.2.0.0
 type FlatTerm = [TermToken]
 
--- | A concrete encoding of @'Enc.Encoding'@ values, one
--- which mirrors the original @'Enc.Encoding'@ type closely.
+-- | A concrete encoding of 'Enc.Encoding' values, one
+-- which mirrors the original 'Enc.Encoding' type closely.
 --
 -- @since 0.2.0.0
 data TermToken
@@ -98,11 +105,11 @@ data TermToken
 
 --------------------------------------------------------------------------------
 
--- | Convert an arbitrary @'Enc.Encoding'@ into a @'FlatTerm'@.
+-- | Convert an arbitrary 'Enc.Encoding' into a 'FlatTerm'.
 --
 -- @since 0.2.0.0
-toFlatTerm :: Encoding -- ^ The input @'Enc.Encoding'@.
-           -> FlatTerm -- ^ The resulting @'FlatTerm'@.
+toFlatTerm :: Encoding -- ^ The input 'Enc.Encoding'.
+           -> FlatTerm -- ^ The resulting 'FlatTerm'.
 toFlatTerm (Encoding tb) = convFlatTerm (tb Enc.TkEnd)
 
 convFlatTerm :: Enc.Tokens -> FlatTerm
@@ -141,17 +148,131 @@ convFlatTerm (Enc.TkFloat16  f  ts) = TkFloat16   f : convFlatTerm ts
 convFlatTerm (Enc.TkFloat32  f  ts) = TkFloat32   f : convFlatTerm ts
 convFlatTerm (Enc.TkFloat64  f  ts) = TkFloat64   f : convFlatTerm ts
 convFlatTerm (Enc.TkBreak       ts) = TkBreak       : convFlatTerm ts
+convFlatTerm (Enc.TkEncoded  bs ts) = decodePreEncoded bs
+                                                   ++ convFlatTerm ts
 convFlatTerm  Enc.TkEnd             = []
 
 --------------------------------------------------------------------------------
 
--- | Given a @'Dec.Decoder'@, decode a @'FlatTerm'@ back into
+decodePreEncoded :: BS.ByteString -> FlatTerm
+decodePreEncoded bs0 =
+    ST.Lazy.runST (provideInput bs0)
+  where
+    provideInput :: BS.ByteString -> ST.Lazy.ST s FlatTerm
+    provideInput bs
+      | BS.null bs = return []
+      | otherwise  = do
+          next <- ST.Lazy.strictToLazyST $ do
+              -- This will always be a 'Partial' here because decodeTermToken
+              -- always starts by requesting initial input. Only decoders that
+              -- fail or return a value without looking at their input can give
+              -- a different initial result.
+              Read.Partial k <- Read.deserialiseIncremental decodeTermToken
+              k (Just bs)
+          collectOutput next
+
+    collectOutput :: Read.IDecode s TermToken -> ST.Lazy.ST s FlatTerm
+    collectOutput (Read.Fail _ _ err) = fail $ "toFlatTerm: encodePreEncoded "
+                                            ++ "used with invalid CBOR: "
+                                            ++ show err
+    collectOutput (Read.Partial    k) = ST.Lazy.strictToLazyST (k Nothing)
+                                        >>= collectOutput
+    collectOutput (Read.Done bs' _ x) = do xs <- provideInput bs'
+                                           return (x : xs)
+
+decodeTermToken :: Decoder s TermToken
+decodeTermToken = do
+    tkty <- peekTokenType
+    case tkty of
+      TypeUInt   -> do w <- decodeWord
+                       return $! fromWord w
+                    where
+                      fromWord :: Word -> TermToken
+                      fromWord w
+                        | w <= fromIntegral (maxBound :: Int)
+                                    = TkInt     (fromIntegral w)
+                        | otherwise = TkInteger (fromIntegral w)
+
+      TypeUInt64 -> do w <- decodeWord64
+                       return $! fromWord64 w
+                    where
+                      fromWord64 w
+                        | w <= fromIntegral (maxBound :: Int)
+                                    = TkInt     (fromIntegral w)
+                        | otherwise = TkInteger (fromIntegral w)
+
+      TypeNInt   -> do w <- decodeNegWord
+                       return $! fromNegWord w
+                    where
+                      fromNegWord w
+                        | w <= fromIntegral (maxBound :: Int)
+                                    = TkInt     (-1 - fromIntegral w)
+                        | otherwise = TkInteger (-1 - fromIntegral w)
+
+      TypeNInt64 -> do w <- decodeNegWord64
+                       return $! fromNegWord64 w
+                    where
+                      fromNegWord64 w
+                        | w <= fromIntegral (maxBound :: Int)
+                                    = TkInt     (-1 - fromIntegral w)
+                        | otherwise = TkInteger (-1 - fromIntegral w)
+
+      TypeInteger -> do !x <- decodeInteger
+                        return (TkInteger x)
+      TypeFloat16 -> do !x <- decodeFloat
+                        return (TkFloat16 x)
+      TypeFloat32 -> do !x <- decodeFloat
+                        return (TkFloat32 x)
+      TypeFloat64 -> do !x <- decodeDouble
+                        return (TkFloat64 x)
+
+      TypeBytes        -> do !x <- decodeBytes
+                             return (TkBytes x)
+      TypeBytesIndef   -> do decodeBytesIndef
+                             return TkBytesBegin
+      TypeString       -> do !x <- decodeString
+                             return (TkString x)
+      TypeStringIndef  -> do decodeStringIndef
+                             return TkStringBegin
+
+      TypeListLen      -> do !x <- decodeListLen
+                             return $! TkListLen (fromIntegral x)
+      TypeListLen64    -> do !x <- decodeListLen
+                             return $! TkListLen (fromIntegral x)
+      TypeListLenIndef -> do decodeListLenIndef
+                             return TkListBegin
+      TypeMapLen       -> do !x <- decodeMapLen
+                             return $! TkMapLen (fromIntegral x)
+      TypeMapLen64     -> do !x <- decodeMapLen
+                             return $! TkMapLen (fromIntegral x)
+      TypeMapLenIndef  -> do decodeMapLenIndef
+                             return TkMapBegin
+
+      TypeTag          -> do !x <- decodeTag
+                             return $! TkTag (fromIntegral x)
+      TypeTag64        -> do !x <- decodeTag
+                             return $! TkTag (fromIntegral x)
+
+      TypeBool    -> do !x <- decodeBool
+                        return (TkBool x)
+      TypeNull    -> do decodeNull
+                        return TkNull
+      TypeSimple  -> do !x <- decodeSimple
+                        return (TkSimple x)
+      TypeBreak   -> do _ <- decodeBreakOr
+                        return TkBreak
+      TypeInvalid -> fail "invalid token encoding"
+
+
+--------------------------------------------------------------------------------
+
+-- | Given a 'Dec.Decoder', decode a 'FlatTerm' back into
 -- an ordinary value, or return an error.
 --
 -- @since 0.2.0.0
 fromFlatTerm :: (forall s. Decoder s a)
-                                -- ^ A @'Dec.Decoder'@ for a serialised value.
-             -> FlatTerm        -- ^ The serialised @'FlatTerm'@.
+                                -- ^ A 'Dec.Decoder' for a serialised value.
+             -> FlatTerm        -- ^ The serialised 'FlatTerm'.
              -> Either String a -- ^ The deserialised value, or an error.
 fromFlatTerm decoder ft =
     runST (getDecodeAction decoder >>= go ft)
@@ -328,7 +449,15 @@ fromFlatTerm decoder ft =
 
     go ts@(tk:_) (PeekTokenType k) = k (tokenTypeOf tk) >>= go ts
     go ts        (PeekTokenType _) = unexpected "peekTokenType" ts
+
+    -- We don't have real bytes so we have to give these two operations
+    -- different interpretations: remaining tokens and just 0 for offsets.
     go ts        (PeekAvailable k) = k (unI# (length ts)) >>= go ts
+#if defined(ARCH_32bit)
+    go ts        (PeekByteOffset k)= k (unI64# 0) >>= go ts
+#else
+    go ts        (PeekByteOffset k)= k 0# >>= go ts
+#endif
 
     go _  (Fail msg) = return $ Left msg
     go [] (Done x)   = return $ Right x
@@ -415,7 +544,7 @@ fromFlatTerm decoder ft =
     unexpected name []      = return $ Left $ name ++ ": unexpected end of input"
     unexpected name (tok:_) = return $ Left $ name ++ ": unexpected token " ++ show tok
 
--- | Map a @'TermToken'@ to the underlying CBOR @'TokenType'@
+-- | Map a 'TermToken' to the underlying CBOR 'TokenType'
 tokenTypeOf :: TermToken -> TokenType
 tokenTypeOf (TkInt n)
     | n >= 0                = TypeUInt
@@ -440,12 +569,12 @@ tokenTypeOf TkFloat64{}     = TypeFloat64
 
 --------------------------------------------------------------------------------
 
--- | Ensure a @'FlatTerm'@ is internally consistent and was created in a valid
+-- | Ensure a 'FlatTerm' is internally consistent and was created in a valid
 -- manner.
 --
 -- @since 0.2.0.0
-validFlatTerm :: FlatTerm -- ^ The input @'FlatTerm'@
-              -> Bool     -- ^ @'True'@ if valid, @'False'@ otherwise.
+validFlatTerm :: FlatTerm -- ^ The input 'FlatTerm'
+              -> Bool     -- ^ 'True' if valid, 'False' otherwise.
 validFlatTerm ts =
    either (const False) (const True) $ do
      ts' <- validateTerm TopLevelSingle ts
@@ -454,7 +583,7 @@ validFlatTerm ts =
        _  -> Left "trailing data"
 
 -- | A data type used for tracking the position we're at
--- as we traverse a @'FlatTerm'@ and make sure it's valid.
+-- as we traverse a 'FlatTerm' and make sure it's valid.
 data Loc = TopLevelSingle
          | TopLevelSequence
          | InString   Int     Loc
@@ -468,7 +597,7 @@ data Loc = TopLevelSingle
          | InTagged   Word64  Loc
   deriving Show
 
--- | Validate an arbitrary @'FlatTerm'@ at an arbitrary location.
+-- | Validate an arbitrary 'FlatTerm' at an arbitrary location.
 validateTerm :: Loc -> FlatTerm -> Either String FlatTerm
 validateTerm _loc (TkInt       _   : ts) = return ts
 validateTerm _loc (TkInteger   _   : ts) = return ts
@@ -562,8 +691,8 @@ maxInt32    = fromIntegral (maxBound :: Int32)
 minInt32    = fromIntegral (minBound :: Int32)
 maxWord32   = fromIntegral (maxBound :: Word32)
 
--- | Do a careful check to ensure an @'Int'@ is in the
--- range of a @'Word32'@.
+-- | Do a careful check to ensure an 'Int' is in the
+-- range of a 'Word32'.
 intIsValidWord32 :: Int -> Bool
 intIsValidWord32 n = b1 && b2
   where
